import os
import subprocess

from dagster import (
    InputDefinition,
    List,
    Nothing,
    Output,
    OutputDefinition,
    Path,
    SolidDefinition,
    check,
    solid,
)

from .configs import define_spark_config
from .types import SparkSolidError
from .utils import parse_spark_config


def create_spark_shell_cmd(solid_config, main_class):
    # Extract parameters from config
    (master_url, deploy_mode, application_jar, spark_conf, application_arguments, spark_home) = [
        solid_config.get(k)
        for k in (
            'master_url',
            'deploy_mode',
            'application_jar',
            'spark_conf',
            'application_arguments',
            'spark_home',
        )
    ]

    # Let the user use env vars in the jar path
    application_jar = os.path.expandvars(application_jar)

    if not os.path.exists(application_jar):
        raise SparkSolidError(
            (
                'Application jar {} does not exist. A valid jar must be '
                'built before running this solid.'.format(application_jar)
            )
        )

    spark_home = spark_home if spark_home else os.environ.get('SPARK_HOME')

    if spark_home is None:
        raise SparkSolidError(
            (
                'No spark home set. You must either pass spark_home in config or '
                'set $SPARK_HOME in your environment (got None).'
            )
        )

    deploy_mode = ['--deploy-mode', '{}'.format(deploy_mode)] if deploy_mode else []

    spark_shell_cmd = (
        ['{}/bin/spark-submit'.format(spark_home), '--class', main_class, '--master', master_url]
        + deploy_mode
        + parse_spark_config(spark_conf)
        + [application_jar]
        + ([application_arguments] if application_arguments else [])
    )
    return spark_shell_cmd


class SparkSolidDefinition(SolidDefinition):
    '''This solid is a generic representation of a parameterized Spark job.

    Parameters:
    name (str): The name of this Spark solid
    main_class (str): The entry point for your application (e.g. org.apache.spark.examples.SparkPi)

    '''

    def __init__(self, name, main_class, spark_outputs=None, description=None):
        name = check.str_param(name, 'name')
        main_class = check.str_param(main_class, 'main_class')
        description = check.opt_str_param(
            description,
            'description',
            'This solid is a generic representation of a parameterized Spark job.',
        )
        spark_outputs = check.opt_list_param(spark_outputs, 'spark_outputs', of_type=str)

        def _spark_compute_fn(context, _):
            '''Define Spark execution.

            This function defines how we'll execute the Spark job and invokes spark-submit.
            '''

            spark_shell_cmd = create_spark_shell_cmd(context.solid_config, main_class)

            context.log.info("Running spark-submit: " + ' '.join(spark_shell_cmd))
            retcode = subprocess.call(' '.join(spark_shell_cmd), shell=True)

            if retcode != 0:
                raise SparkSolidError('Spark job failed. Please consult your logs.')

            yield Output(spark_outputs, 'paths')

        super(SparkSolidDefinition, self).__init__(
            name=name,
            description=description,
            input_defs=[InputDefinition('spark_inputs', List[Path])],
            output_defs=[OutputDefinition(dagster_type=List[Path], name='paths')],
            compute_fn=_spark_compute_fn,
            config=define_spark_config(),
            tags={'kind': 'spark', 'main_class': main_class},
        )


def create_spark_solid(name, main_class, description=None):
    check.str_param(name, 'name')
    check.str_param(main_class, 'main_class')
    check.opt_str_param(description, 'description', 'A parameterized Spark job.')

    @solid(
        name=name,
        description=description,
        config=define_spark_config(),
        input_defs=[InputDefinition('start', Nothing)],
        output_defs=[OutputDefinition(Nothing)],
        tags={'kind': 'spark', 'main_class': main_class},
        required_resource_keys={'spark'},
    )
    def spark_solid(context):  # pylint: disable=unused-argument
        context.resources.spark.run_spark_job(context.solid_config, main_class)

    return spark_solid



import os

import pytest
import yaml
from dagster_spark import create_spark_solid, spark_resource

from dagster import ModeDefinition, execute_pipeline, pipeline

CONFIG = '''
solids:
  first_pi:
    config:
      master_url: "local[2]"
      deploy_mode: "client"
      spark_conf:
        spark:
          app:
            name: "first_pi"
      application_jar: {jar_path}
      application_arguments: '10'

  second_pi:
    config:
      master_url: "local[2]"
      deploy_mode: "client"
      spark_conf:
        spark:
          app:
            name: "second_pi"
      application_jar: {jar_path}
      application_arguments: '10'

  third_pi:
    config:
      master_url: "local[2]"
      deploy_mode: "client"
      spark_conf:
        spark:
          app:
            name: "third_pi"
      application_jar: {jar_path}
      application_arguments: '10'
'''


@pytest.mark.skip('for local testing only, we don\'t have $SPARK_HOME on buildkite yet')
def test_multiple_jobs():
    @pipeline(mode_defs=[ModeDefinition(resource_defs={'spark': spark_resource})])
    def pipe():
        for solid_name in ['first_pi', 'second_pi', 'third_pi']:
            create_spark_solid(solid_name, main_class='org.apache.spark.examples.SparkPi')()

    # Find SPARK_HOME to get to spark examples jar
    base_path = os.path.expandvars('${SPARK_HOME}/examples/jars/')
    jar_path = None
    for fname in os.listdir(base_path):
        if fname.startswith('spark-examples'):
            jar_path = os.path.join(base_path, fname)

    result = execute_pipeline(pipe, yaml.safe_load(CONFIG.format(jar_path=jar_path)))
    assert result.success



from .configs import define_spark_config
from .resources import spark_resource
from .solids import SparkSolidDefinition, create_spark_solid
from .types import SparkSolidError
from .utils import construct_spark_shell_command

__all__ = [
    'construct_spark_shell_command',
    'create_spark_solid',
    'define_spark_config',
    'spark_resource',
    'SparkSolidDefinition',
    'SparkSolidError',
]



from dagster_spark.utils import parse_spark_config


def test_parse_spark_config():
    test = {
        'spark': {
            'app': {'name': 'foo'},
            'driver': {'blockManager': {}},
            'executor': {'pyspark': {}, 'logs': {'rolling': {'time': {}}}},
            'local': {},
            'submit': {},
            'log': {},
            'executorEnv': {},
            'redaction': {},
            'python': {'profile': {}, 'worker': {}},
            'files': {},
            'jars': {},
            'pyspark': {'driver': {}},
            'reducer': {},
            'shuffle': {
                'file': {},
                'io': {},
                'service': {'index': {'cache': {}}},
                'sort': {},
                'spill': {},
                'registration': {},
            },
            'eventLog': {'logBlockUpdates': {}, 'longForm': {}, 'buffer': {}},
            'ui': {'dagGraph': {}, 'liveUpdate': {}},
            'worker': {'ui': {}},
            'sql': {'ui': {}},
            'streaming': {
                'ui': {},
                'backpressure': {},
                'receiver': {'writeAheadLog': {}},
                'kafka': {},
                'driver': {'writeAheadLog': {}},
            },
            'broadcast': {},
            'io': {'compression': {'lz4': {}, 'snappy': {}, 'zstd': {}}},
            'kryo': {},
            'kryoserializer': {'buffer': {}},
            'rdd': {},
            'serializer': {},
            'memory': {'offHeap': {}},
            'storage': {'replication': {}},
            'cleaner': {'periodicGC': {}, 'referenceTracking': {'blocking': {}}},
            'default': {},
            'hadoop': {'mapreduce': {'fileoutputcommitter': {'algorithm': {}}}},
            'rpc': {'message': {}, 'retry': {}},
            'blockManager': {},
            'network': {},
            'port': {},
            'core': {'connection': {'ack': {'wait': {}}}},
            'cores': {'max': '3'},
            'locality': {'wait': {}},
            'scheduler': {'revive': {}, 'listenerbus': {'eventqueue': {}}},
            'blacklist': {'task': {}, 'stage': {}, 'application': {'fetchFailure': {}}},
            'speculation': {},
            'task': {'reaper': {}},
            'stage': {},
            'dynamicAllocation': {},
            'r': {'driver': {}, 'shell': {}},
            'graphx': {'pregel': {}},
            'deploy': {'zookeeper': {}},
        }
    }
    parsed_config = parse_spark_config(test)

    # Iteration order isn't preserved across python versions
    assert set([('--conf', 'spark.app.name=foo'), ('--conf', 'spark.cores.max=3')]) == set(
        [(parsed_config[i], parsed_config[i + 1]) for i in range(0, len(parsed_config), 2)]
    )



'''Spark Configuration

In this file we define the key configuration parameters for submitting Spark jobs. Spark can be run
in a variety of deployment contexts. See the Spark documentation at
https://spark.apache.org/docs/latest/submitting-applications.html for a more in-depth summary of
Spark deployment contexts and configuration.
'''
from dagster import Field

from .configs_spark import spark_config
from .types import SparkDeployMode


def define_spark_config():
    '''Spark configuration.

    See the Spark documentation for reference:
        https://spark.apache.org/docs/latest/submitting-applications.html
    '''

    master_url = Field(
        str,
        description='The master URL for the cluster (e.g. spark://23.195.26.187:7077)',
        is_required=True,
    )

    deploy_mode = Field(
        SparkDeployMode,
        description='''Whether to deploy your driver on the worker nodes (cluster) or locally as an
        external client (client) (default: client). A common deployment strategy is to submit your
        application from a gateway machine that is physically co-located with your worker machines
        (e.g. Master node in a standalone EC2 cluster). In this setup, client mode is appropriate.
        In client mode, the driver is launched directly within the spark-submit process which acts
        as a client to the cluster. The input and output of the application is attached to the
        console. Thus, this mode is especially suitable for applications that involve the REPL (e.g.
        Spark shell).''',
        is_required=False,
    )

    application_jar = Field(
        str,
        description='''Path to a bundled jar including your application and all
                        dependencies. The URL must be globally visible inside of your cluster, for
                        instance, an hdfs:// path or a file:// path that is present on all nodes.
                        ''',
        is_required=True,
    )

    application_arguments = Field(
        str,
        description='Arguments passed to the main method of your main class, if any',
        is_required=False,
    )

    spark_home = Field(
        str,
        description='The path to your spark installation. Defaults to $SPARK_HOME at runtime if not provided.',
        is_required=False,
    )

    return {
        'master_url': master_url,
        'deploy_mode': deploy_mode,
        'application_jar': application_jar,
        'spark_conf': spark_config(),
        'spark_home': spark_home,
        'application_arguments': application_arguments,
    }
